/* 
 * Copyright (c)2007, Dustin Spicuzza
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * The name of the Dustin Spicuzza may not be used to endorse or promote 
 *       products derived from this software without specific prior written 
 *       permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Dustin Spicuzza ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Dustin Spicuzza BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This library extends and supplements the Cypress SuiteUSB .NET C# library
 * 
 * This is not distributed by or affiliated with Cypress or SuiteUSB .NET, and is not
 * supported by them. You need their library to make this work, of course.
 * 
 * Heavily influenced by http://msdn.microsoft.com/msdnmag/issues/07/06/concurrentaffairs/default.aspx
 */

using System;
using System.Threading;
using System.Runtime.InteropServices;
using CyUSB;

namespace CyUSB.Async {

	/// <summary>
	/// Wrapper for the CyUSBDevice that implements .NET-style asynchronous patterns, instead
	/// of the annoying interface that Cypress provides. You pass it a CyUSBEndPoint at construction,
	/// and it will provide nice .NET style asynchronous calls for you to use, which work pretty
	/// nicely. :)
	/// 
	/// It is important to note that this DOES NOT replace the CyUSBDevice class, it is really
	/// just a nice supplement for it. 
	/// </summary>
	public sealed partial class CyUSBAsyncWrapper {

		CyUSBEndPoint endPoint = null;
		IOCompletionCallback transferCallback;


		private CyUSBAsyncWrapper(){}

		public unsafe CyUSBAsyncWrapper(CyUSBEndPoint endPoint) {

			transferCallback = new IOCompletionCallback(TransferCallback);

			this.endPoint = endPoint;

			// don't care about this warning... we have to use an IntPtr, its all we have
#pragma warning disable 0618
			try {
				ThreadPool.BindHandle(endPoint.hDevice);
			} catch (ApplicationException){
				// don't care about this exception
				// it will occur if we try to bind the handle twice, which.. doesn't
				// matter. probably. We'll see how that goes. 
			}
#pragma warning restore 0618

		}

		#region Another level of indirection on the endpoint library

		/// <summary>
		/// Aborts a transfer on the endpoint
		/// </summary>
		public bool Abort() {
			return endPoint.Abort();
		}


		/// <summary>
		/// Resets the endpoint
		/// </summary>
		public bool Reset() {
			return endPoint.Reset();
		}


		#endregion

		/// <summary>
		/// Performs a blocking transfer to the device
		/// </summary>
		/// <param name="buffer">Buffer to read/write</param>
		/// <param name="len">Length to read/write</param>
		public int Transfer(byte[] buffer, int len) {


			// check for stupid mistakes first
			if ( buffer == null )
				throw new ArgumentNullException("buffer");

			if ( buffer.Length < len )
				throw new ArgumentException("Length exceeds the size of the buffer", "len");

			// this doesn't always work, however
			if ( endPoint.hDevice == CyConst.INVALID_HANDLE )
				throw new ArgumentException("The endpoint is no longer valid");

			// necessary parameters for call
			uint deviceControlCode = 0x22004b;

			bool ret;
			int bytesReturned = 0;

			byte[] inputBuffer = new byte[CyConst.SINGLE_XFER_LEN];

			unsafe {

				// pin this down
				fixed ( byte* singleRef = inputBuffer ) {

					// create this thing
					SINGLE_TRANSFER* single_transferPtr = (SINGLE_TRANSFER*)singleRef;
					single_transferPtr->WaitForever = 0;
					single_transferPtr->ucEndpointAddress = endPoint.Address;
					single_transferPtr->IsoPacketLength = 0;
					single_transferPtr->BufferOffset = 0;
					single_transferPtr->BufferLength = 0;

					// finally execute the ioctl
					ret = NativeMethods.DeviceIoControl(endPoint.hDevice, deviceControlCode, single_transferPtr, CyConst.SINGLE_XFER_LEN, buffer, len, out bytesReturned, null);

				}
			}

			// success
			if ( ret )
				return bytesReturned;

			// throw an exception, let it figure out what happened
			throw new USBException();
			
		}
	

		/// <summary>
		/// Begins an asynchronous transfer on the endpoint
		/// </summary>
		/// <param name="buffer">Buffer to read/write</param>
		/// <param name="len">Length to read/write</param>
		public IAsyncResult BeginTransfer(byte[] buffer, int len, AsyncCallback userCallback, object asyncState) {


			// check for stupid mistakes first
			if ( buffer == null )
				throw new ArgumentNullException("buffer");

			if ( buffer.Length < len )
				throw new ArgumentException("Length exceeds the size of the buffer", "len");

			// this doesn't always work, however
			if ( endPoint.hDevice == CyConst.INVALID_HANDLE)
				throw new ArgumentException("The endpoint is no longer valid");

			// create a new asyncresult
			CyUSBAsyncResult result = new CyUSBAsyncResult();

			// populate its fields
			result.callback = userCallback;
			result.asyncState = asyncState;

			// create the overlapped io struct stuff
			result.overlapped = new Overlapped(0,0,IntPtr.Zero, result);

			// necessary parameters for call
			uint deviceControlCode = 0x22004b;

			bool ret;
			int bytesReturned = 0;

			byte [] inputBuffer = new byte[CyConst.SINGLE_XFER_LEN];

			unsafe{

				// pack the buffer and result into the overlapped thing
				NativeOverlapped * ov = result.overlapped.Pack(transferCallback, buffer);

				// pin this down
				fixed ( byte * singleRef = inputBuffer ) {

					// create this thing
					SINGLE_TRANSFER* single_transferPtr = (SINGLE_TRANSFER*)singleRef;
					single_transferPtr->WaitForever = 0;
					single_transferPtr->ucEndpointAddress = endPoint.Address;
					single_transferPtr->IsoPacketLength = 0;
					single_transferPtr->BufferOffset = 0;
					single_transferPtr->BufferLength = 0;

					// finally execute the ioctl
					ret = NativeMethods.DeviceIoControl(endPoint.hDevice, deviceControlCode, single_transferPtr, CyConst.SINGLE_XFER_LEN, buffer, len, out bytesReturned, ov);
        
				}
			}
		
			// success
			if (ret) 
				return result;

			result.errorCode = Marshal.GetLastWin32Error();

			// io pending.. we expect this to happen
			if ( result.errorCode == 997 ) 
				return result;

			// throw an exception, let it figure out what happened
			throw new USBException(result.errorCode);
		}


		// io completion port callback thing
		private unsafe void TransferCallback(uint errorCode, uint numBytes, NativeOverlapped* pOVERLAP) {

			// get the data we need
			Overlapped overlapped = Overlapped.Unpack(pOVERLAP);
			CyUSBAsyncResult result = overlapped.AsyncResult as CyUSBAsyncResult;

			// free this, don't need it anymore
			Overlapped.Free(pOVERLAP);

            result.errorCode = (int)errorCode;

			// keep this information around, we need to return it later
			result.bytesTransferred = numBytes;

			// release anyone waiting on this
			result.asyncWaitHandle.Set();

			// if there is a callback, then call it and we're done
			if ( result.callback != null )
				result.callback(result);
		}

		/// <summary>
		/// Ends an asynchronous transfer on the endpoint. Blocks until it finishes. 
		/// </summary>
		public int EndTransfer(IAsyncResult asyncResult) {

			// simple error checking, don't do anything crazy
			if ( asyncResult == null )
				throw new ArgumentNullException("asyncResult");

			// cast the result so we can get its internals
			CyUSBAsyncResult result = asyncResult as CyUSBAsyncResult;

			// wait for the operation to complete -- generally, this shouldn't be necessary
			try {
				result.AsyncWaitHandle.WaitOne();
			} finally {
				// close the handle, we don't need it anymore
				result.AsyncWaitHandle.Close();
			}

			// and... we're done
			result.isCompleted = true;

			// is there an error code? 
			if ( result.errorCode != 0 )
				throw new USBException(result.errorCode);
			
			// nope, return the result
			return (int)result.bytesTransferred;
		}

	}

	public sealed class CyUSBAsyncResult : IAsyncResult  {

		internal bool isCompleted = false;

		internal Overlapped overlapped;
		internal object asyncState;
		internal ManualResetEvent asyncWaitHandle = new ManualResetEvent(false);
		internal AsyncCallback callback;

		internal uint bytesTransferred = 0;

		internal int errorCode = 0;

		#region IAsyncResult Members

		/// <summary>
		/// User-specified state object
		/// </summary>
		public object AsyncState {
			get { return asyncState; }
		}

		public WaitHandle AsyncWaitHandle {
			get { return asyncWaitHandle; }
		}

		/// <summary>
		/// This always returns false
		/// </summary>
		public bool CompletedSynchronously {
			get { return false; }
		}

		public bool IsCompleted {
			get { return isCompleted; }
		}

		#endregion

	}

	/// <summary>
	/// Represents errors that occur during USB communication
	/// </summary>
	[Serializable]
	public sealed class USBException : Exception {

		USBErrorCode usbError;
		string message = null;

		public USBErrorCode UsbError {
			get { return usbError; }
		}

		/// <summary>
		/// Automatically determines the error condition from Marshal.GetLastWin32Error
		/// </summary>
		public USBException() {
			init(Marshal.GetLastWin32Error());
		}

		/// <summary>
		/// Pass this the error code, and it will automatically determine an error message
		/// </summary>
		/// <param name="ErrorCode"></param>
		public USBException(int ErrorCode) {

			init(ErrorCode);
		}

		private void init(int errorCode) {

			HResult = ErrorCode;

			switch ( HResult ) {
				case 5:
					usbError = USBErrorCode.AccessDenied;
					message = "Access was denied when transferring with the USB device";
					break;

				case 6:
					usbError = USBErrorCode.InvalidDevice;
					message = "The USB transfer failed because the USB device is no longer attached";
					break;

				case -2146233088:
					usbError = USBErrorCode.InvalidDevice;
					message = "The USB transfer failed because the USB device is no longer attached";
					break;

				default:
					usbError = USBErrorCode.TransferFailed;
					message = "The USB transfer failed because of an unknown error (Code 0x" + HResult.ToString("X") + ")";
					break;
			}
		}

		/// <summary>
		/// Gets message that describes the current exception
		/// </summary>
		public override string Message {
			get {
				if (message != null)
					return message;
				
				return base.Message;
			}
		}

		/// <summary>
		/// An error code that describes the error condition
		/// </summary>
		public int ErrorCode {
			get {
				return HResult;
			}
		}
	}

	/// <summary>
	/// Defines common USB error codes
	/// </summary>
	public enum USBErrorCode {

		/// <summary>
		/// The transfer failed. Check the errorcode to see more information
		/// </summary>
		TransferFailed,

		/// <summary>
		/// Access denied when trying to use the IOCTL
		/// </summary>
		AccessDenied,

		/// <summary>
		/// Invalid device. This will generally occur when you deteched the USB device
		/// </summary>
		InvalidDevice
	}
}
